import random
import math

import pymunk

from bitflags import cfg
from bitflags.rgb import RGB
from bitflags.data import img
from bitflags.utils import angle_between, dir_from_degrees
from bitflags.fx import PointExplosion


enemy_images = {
    '001': img('enemy-001.png'),
    '010': img('enemy-010.png'),
    '100': img('enemy-100.png'),

    '011': img('enemy-011.png'),
    '110': img('enemy-110.png'),
    '101': img('enemy-101.png'),

    '111': img('enemy-111.png'),
}


class Enemy(RGB):
    ACTION_MOVE = 'move'
    ACTION_MOVE_REVERSE = 'move_reverse'
    ACTION_FIND_PLAYER = 'find'
    ACTION_FIND_PLAYER_REVERSE = 'find_reverse'

    TIMERS = {
        ACTION_MOVE: 2.5,
        ACTION_MOVE_REVERSE: 2.5,
        ACTION_FIND_PLAYER: 1.0,
        ACTION_FIND_PLAYER_REVERSE: 1.0,
    }

    def __init__(self, r, g, b, pos, rot, on_killed):
        super(Enemy, self).__init__(r, g, b)
        self.on_killed_for_room = on_killed
        self.on_killed_for_scene = None
        self.on_shoot = None
        self.rotation = rot
        self.position = pos
        self.player = None

        self.direction = dir_from_degrees(self.rotation)

        self.timer = 0
        self.action = self.ACTION_MOVE
        self.reset_timer()

        self.cbody = pymunk.Body(1, 1)
        self.cshape = pymunk.Circle(self.cbody, 26.0)
        self.cshape.collision_type = cfg.COLL_TA
        self.cshape.on_hit = self.die
        self.cshape.rgb = self.rgb()

        self.cbody.position = self.position
        self.cbody.angle = math.radians(-self.rotation)
        self.initial_angle = self.cbody.angle

    def die(self, space):
        space.remove(self.cshape)
        self.on_killed_for_room(self)
        self.on_killed_for_scene()

        self.opacity = 0.0
        self.add(PointExplosion((0, 0), self.kill))

    def reset_timer(self):
        self.timer = self.TIMERS[self.action]

    def dec_timer(self, dt):
        self.timer -= dt
        if self.timer <= 0.0:
            method_name = 'end_{0}'.format(self.action)
            getattr(self, method_name)()
            self.reset_timer()

    def get_image_map(self):
        return enemy_images

    def graphics(self, dt):
        self.position = self.cbody.position
        self.rotation = -math.degrees(self.cbody.angle)

    def physics(self, dt):
        self.dec_timer(dt)

        method_name = 'do_{0}_action'.format(self.action)
        getattr(self, method_name)(dt)

    def do_move_action(self, dt):
        self.cbody.position += self.direction * dt * cfg.ENEMY_SPEED

    def end_move(self):
        if random.random() < cfg.ENEMY_CHANCE_TO_SHOOT:
            self.action = self.ACTION_FIND_PLAYER
        else:
            self.cbody.angle = self.initial_angle + math.pi
            self.action = self.ACTION_MOVE_REVERSE

    def do_move_reverse_action(self, dt):
        self.cbody.position -= self.direction * dt * cfg.ENEMY_SPEED

    def end_move_reverse(self):
        if random.random() < cfg.ENEMY_CHANCE_TO_SHOOT:
            self.action = self.ACTION_FIND_PLAYER_REVERSE
        else:
            self.cbody.angle = self.initial_angle
            self.action = self.ACTION_MOVE

    def do_find_action(self, dt):
        self.find_player(dt)

    def find_player(self, dt):
        angle = angle_between(self.cbody, self.player.cbody)
        self.cbody.angle += angle * cfg.ENEMY_ROT_MULTIPLY

    def end_find(self):
        self.cbody.angle = self.initial_angle + math.pi
        self.action = self.ACTION_MOVE_REVERSE
        self.try_to_shoot()

    def do_find_reverse_action(self, dt):
        self.find_player(dt)

    def end_find_reverse(self):
        self.cbody.angle = self.initial_angle
        self.action = self.ACTION_MOVE
        self.try_to_shoot()

    def try_to_shoot(self):
        self.on_shoot(self)
