import math

import pymunk
from pymunk.vec2d import Vec2d

from cocos.sprite import Sprite

from bitflags import cfg
from bitflags.data import img
from bitflags.constraint import Constraint
from bitflags.utils import angle_between, dir_from_degrees


class Player(Sprite):
    def __init__(self, bits, sensor):
        self.is_sensor_on = False

        self.bits = bits
        self.sensor = sensor

        super(Player, self).__init__(img('player.png'))
        self.position = cfg.PL_INIT_POS
        self.add(self.sensor, z=-1)

        self.init_physics()

        self.pos_dir = Vec2d(0.0, 0.0)
        self.rot_dir = 0.0
        self.walls = {}

    def init_physics(self):
        self.cbody = pymunk.Body(1, 1)
        self.cbody.position = self.position
        self.cbody.angle = self.rotation
        self.cshape = pymunk.Poly(self.cbody,
                                  self.get_cshape_points())
        self.cshape.collision_type = cfg.COLL_PL

    def get_cshape_points(self):
        return [
            (-30, -25),
            (-30, +20),

            (-20, +24),
            (-10, +28),
            (+10, +28),
            (+20, +24),

            (+30, +20),
            (+30, -25),
        ]

    def physics(self, dt):
        self.evaluate_bits()

        self.cbody.angle += self.rot_dir * dt * cfg.PL_ROT_SPEED
        self.cbody.position += self.pos_dir * dt * cfg.PL_POS_SPEED

        if self.walls:
            for c in self.walls.values():
                self.cbody.position = c.contain(self.cbody.position)

        self.sensor.fallow_player(self)

    def graphics(self, dt):
        self.position = self.cbody.position
        self.rotation = -math.degrees(self.cbody.angle)

    def evaluate_bits(self):
        self.is_sensor_on = False
        self.rgb = [0, 0, 0]

        self.pos_dir = Vec2d(0, 0)
        self.rot_dir = 0.0

        for bit in (b for b in self.bits if b.is_on()):
            getattr(self, 'bit_{0}_job'.format(bit.number))()

        self.pos_dir = self.pos_dir.normalized()

        if self.is_sensor_on and any(self.rgb):
            target = self.sensor.get_target()
            if target is not None:
                self.fallow_target(target)

    def fallow_target(self, target):
        angle = angle_between(self.cbody, target.body)

        if self.cbody.position.get_distance(target.body.position) < 200.0:
            self.pos_dir = Vec2d(0, 0)
        else:
            self.bit_1_job()

        self.cbody.angle += angle * cfg.PL_ROT_MULTIPLY
        self.rot_dir = 0.0

    def contain_within_room(self):
        padding = cfg.PL_RADIUS + cfg.ROOM_PADDING
        x, y = self.cbody.position
        x = max(padding, min(x, cfg.W - padding))
        y = max(padding, min(y, cfg.H - padding))
        self.cbody.position = Vec2d(x, y)

    def on_hit_wall_begin(self, space, arbiter):
        segment = arbiter.shapes[1]
        self.walls[segment] = Constraint(self, segment)
        return 0

    def on_hit_wall_end(self, space, arbiter):
        del(self.walls[arbiter.shapes[1]])
        return 0

    def bit_1_job(self):
        self.pos_dir = dir_from_degrees(self.rotation)

    def bit_2_job(self):
        self.rot_dir += 1.0

    def bit_3_job(self):
        self.rot_dir -= 1.0

    def bit_4_job(self):
        self.is_sensor_on = True

    def bit_5_job(self):
        self.rgb[0] = 1

    def bit_6_job(self):
        self.rgb[1] = 1

    def bit_7_job(self):
        self.rgb[2] = 1

    def bit_8_job(self):
        pass
