import math

import pymunk
from pymunk.vec2d import Vec2d

from cocos.text import Label
from cocos.sprite import Sprite

from bitflags import cfg
from bitflags.data import img
from bitflags.loot import Loot
from bitflags.gate import Gate
from bitflags.enemy import Enemy


class Room(Sprite):
    def __init__(self, index, exits):
        self.index = index

        super(Room, self).__init__(img('borders.png'))
        self.image_anchor = (0, 0)

        self.add(Label(
            u'{0:05}'.format(index), cfg.CENTER,
            color=cfg.ROOM_LABEL_COLOR,
            font_name='Munro',
            font_size=cfg.ROOM_LABEL_SIZE,
            anchor_x='center', anchor_y='center'
        ))

        self.cshapes = []
        for p1, p2 in self.get_verts(exits.keys()):
            shape = pymunk.Segment(pymunk.Body(), p1, p2, 4)
            shape.collision_type = cfg.COLL_RM
            self.cshapes.append(shape)

        for exit_type, next_index in exits.iteritems():
            self.create_exit(exit_type, next_index)

        self.enemies = []

        self.add_custom_items()

    def add_custom_items(self):
        pass

    def graphics(self, dt):
        for enemy in self.enemies:
            enemy.graphics(dt)

    def physics(self, dt):
        for enemy in self.enemies:
            enemy.physics(dt)

    def create_exit(self, exit_type, next_index):
        exit = Exit(exit_type, next_index)
        self.add(exit)
        self.cshapes.append(exit.cshape)

    def create_loot(self, loot_pos, bit_index):
        loot = Loot(loot_pos, bit_index)
        self.add(loot)
        self.cshapes.append(loot.cshape)

    def create_gates(self, entry_type, r, g, b):
        if entry_type == Exit.N:
            pos = Vec2d(cfg.HW, cfg.H - cfg.GATE_DELTA)
            angle = 0.0
        elif entry_type == Exit.S:
            pos = Vec2d(cfg.HW, cfg.GATE_DELTA)
            angle = 180.0
        elif entry_type == Exit.W:
            pos = Vec2d(cfg.GATE_DELTA, cfg.HH)
            angle = -90.0
        elif entry_type == Exit.E:
            pos = Vec2d(cfg.W - cfg.GATE_DELTA, cfg.HH)
            angle = 90.0

        gate = Gate(r, g, b, pos, angle, self.on_gate_remove_cshape)
        self.add(gate)
        self.cshapes.extend(gate.cshapes)

    def on_gate_remove_cshape(self, cshape):
        self.cshapes.remove(cshape)

    def create_enemy(self, r, g, b, pos, rot):
        enemy = Enemy(r, g, b, pos, rot, self.on_enemy_die)
        self.enemies.append(enemy)
        self.add(enemy)
        self.cshapes.append(enemy.cshape)

    def on_enemy_die(self, enemy):
        self.enemies.remove(enemy)
        self.cshapes.remove(enemy.cshape)

    def place_player(self, player, entry_type):
        if entry_type == Exit.N:
            player.cbody.position = Vec2d(cfg.HW, cfg.H)
            player.cbody.angle = math.radians(180.0)
        elif entry_type == Exit.W:
            player.cbody.position = Vec2d(0, cfg.HH)
            player.cbody.angle = math.radians(270.0)
        elif entry_type == Exit.S:
            player.cbody.position = Vec2d(cfg.HW, 0)
            player.cbody.angle = math.radians(0.0)
        elif entry_type == Exit.E:
            player.cbody.position = Vec2d(cfg.W, cfg.HH)
            player.cbody.angle = math.radians(90.0)

    def update_enemies(self, player, on_shoot, on_die):
        for enemy in self.enemies:
            enemy.player = player
            enemy.on_shoot = on_shoot
            enemy.on_killed_for_scene = on_die

    def get_verts(self, exits):
        cx, cy = cfg.CENTER
        rx = cfg.HW - cfg.ROOM_PADDING
        ry = cfg.HH - cfg.ROOM_PADDING
        w = cfg.EXIT_W / 2.0

        return (
            self.north_verts(exits, cx, cy, rx, ry, w) +
            self.east_verts(exits, cx, cy, rx, ry, w) +
            self.west_verts(exits, cx, cy, rx, ry, w) +
            self.south_verts(exits, cx, cy, rx, ry, w)
        )

    def north_verts(self, exits, cx, cy, rx, ry, w):
        if Exit.N in exits:
            return [
                ((cx - rx, cy + ry), (cx - w, cy + ry)),
                ((cx + w, cy + ry), (cx + rx, cy + ry))
            ]
        return [
            ((cx - rx, cy + ry), (cx + rx, cy + ry)),
        ]

    def east_verts(self, exits, cx, cy, rx, ry, w):
        if Exit.E in exits:
            return [
                ((cx + rx, cy - ry), (cx + rx, cy - w)),
                ((cx + rx, cy + w), (cx + rx, cy + ry))
            ]
        return [
            ((cx + rx, cy - ry), (cx + rx, cy + ry))
        ]

    def west_verts(self, exits, cx, cy, rx, ry, w):
        if Exit.W in exits:
            return [
                ((cx - rx, cy - ry), (cx - rx, cy - w)),
                ((cx - rx, cy + w), (cx - rx, cy + ry))
            ]
        return [
            ((cx - rx, cy - ry), (cx - rx, cy + ry))
        ]

    def south_verts(self, exits, cx, cy, rx, ry, w):
        if Exit.S in exits:
            return [
                ((cx - rx, cy - ry), (cx - w, cy - ry)),
                ((cx + w, cy - ry), (cx + rx, cy - ry)),
            ]
        return [
            ((cx - rx, cy - ry), (cx + rx, cy - ry)),
        ]


class Exit(Sprite):
    N = 1
    E = 2
    S = 3
    W = 4

    mirror_map = {
        N: S,
        E: W,
        S: N,
        W: E
    }

    def __init__(self, exit_type, next_index):
        super(Exit, self).__init__(img('exit.png'))

        w, delta = cfg.EXIT_W / 2.0, cfg.EXIT_DELTA

        if exit_type == Exit.N:
            self.rotation = 0.0
            self.position = (cfg.HW, cfg.H - cfg.ROOM_PADDING)
            a, b = (-w, delta), (w, delta)
        elif exit_type == Exit.S:
            self.rotation = 180.0
            self.position = (cfg.HW, cfg.ROOM_PADDING)
            a, b = (-w, -delta), (w, -delta)
        elif exit_type == Exit.E:
            self.rotation = 90.0
            self.position = (cfg.W - cfg.ROOM_PADDING, cfg.HH)
            a, b = (delta, -w), (delta, w)
        elif exit_type == Exit.W:
            self.rotation = 270.0
            self.position = (cfg.ROOM_PADDING, cfg.HH)
            a, b = (-delta, -w), (-delta, w)

        self.cshape = pymunk.Segment(
            pymunk.Body(), a, b, 4
        )
        self.cshape.body.position = self.position
        self.cshape.collision_type = cfg.COLL_EX
        self.cshape.room_index = next_index
        self.cshape.entry_type = self.mirror_map[exit_type]
