import logging

from cocos.scene import Scene

import pymunk

from bitflags import cfg
from bitflags.layers import (BackgroundLayer,
                             GameObjectsLayer,
                             UILayer,
                             StartLayer,
                             EndLayer)
from cocos.director import director
from bitflags.player import Player
from bitflags.bit import Bit, SensorBit, RBit, GBit, BBit, ShootBit
from bitflags.rooms import init_rooms
from bitflags.bullet import Bullet
from bitflags.sensor import Sensor


class EndScene(Scene):
    def __init__(self):
        super(EndScene, self).__init__(EndLayer())


class StartScene(Scene):
    def __init__(self, next_scene):
        super(StartScene, self).__init__(StartLayer(next_scene))


class GameScene(Scene):
    def __init__(self):
        self.sensor = Sensor(0, 0, 0)
        self.enemies_killed = 0

        self.bits = [
            Bit(1, cfg.BIT_FALSE),
            Bit(2),
            Bit(3),

            SensorBit(self.sensor, 4),
            RBit(self.sensor, 5),
            GBit(self.sensor, 6),
            BBit(self.sensor, 7),

            ShootBit(self.on_shooting_begin, self.on_shooting_end, 8),
        ]

        if cfg.TURN_ON_BITS:
            for b in self.bits:
                b.activate()

        self.player = Player(self.bits, self.sensor)
        self.rooms = init_rooms()
        self.active_room = None
        self.set_room(0)

        self.bullets = []

        self.init_physics()

        self.background = BackgroundLayer(self.active_room)
        self.ui_layer = UILayer(self.bits)
        self.game_objects = GameObjectsLayer(self.player)

        super(GameScene, self).__init__(
            self.background,
            self.game_objects,
            self.ui_layer
        )
        self.schedule(self.loop)

    def init_physics(self):
        self.space = pymunk.Space()
        self.space.add(self.sensor.cshape)
        self.space.add(self.player.cshape)
        self.space.add(*self.active_room.cshapes)

        self.space.add_collision_handler(cfg.COLL_PL, cfg.COLL_RM,
                                         begin=self.player.on_hit_wall_begin,
                                         separate=self.player.on_hit_wall_end)
        self.space.add_collision_handler(cfg.COLL_PL, cfg.COLL_EX,
                                         begin=self.on_next_room)
        self.space.add_collision_handler(cfg.COLL_PL, cfg.COLL_LT,
                                         begin=self.on_found_a_bit)
        self.space.add_collision_handler(cfg.COLL_SE, cfg.COLL_TA,
                                         begin=self.sensor.on_sees_target,
                                         separate=self.sensor.on_looses_target)
        self.space.add_collision_handler(cfg.COLL_BU, cfg.COLL_TA,
                                         begin=self.on_bullet_hit_target)
        self.space.add_collision_handler(cfg.COLL_BU, cfg.COLL_PL,
                                         begin=self.on_player_die)

    def on_player_die(self, space, arbiter):
        next_index = max(0, self.active_room.index - 1)
        if next_index == self.active_room.index:
            logging.warning("Enemy is not suppose to be in zero room")
        else:
            self.enter_room(next_index)

        self.player.cbody.position = cfg.PL_INIT_POS
        self.player.cbody.angle = 0.0

        for b in self.bits:
            if b.status == cfg.BIT_TRUE:
                b.update_image()
        self.sensor.r, self.sensor.g, self.sensor.b = 0, 0, 0
        self.sensor.turn_off()
        self.sensor.update_image()
        self.unschedule(self.player_shoot)
        return 0

    def loop(self, dt):
        self.player.graphics(dt)
        for b in self.bullets:
            b.graphics(dt)
        self.active_room.graphics(dt)

        self.space.step(dt)

        self.player.physics(dt)
        for b in self.bullets:
            b.physics(dt)
        self.active_room.physics(dt)

    def on_next_room(self, space, arbiter):
        exit = arbiter.shapes[1]
        self.enter_room(exit.room_index, exit.entry_type)
        return 0

    def enter_room(self, index, entry_type=None):
        self.space.remove(*self.active_room.cshapes)

        self.set_room(index)

        self.space.add(*self.active_room.cshapes)

        self.background.replace_room(self.active_room)

        for b in self.bullets:
            self.space.remove(b.cshape)
            b.kill()
        self.bullets = []
        self.sensor.targets = []

        if entry_type is not None:
            self.active_room.place_player(self.player, entry_type)

    def set_room(self, index):
        self.active_room = self.rooms[index]
        self.active_room.update_enemies(self.player,
                                        self.on_enemy_shoot,
                                        self.on_enemy_die)

    def on_enemy_die(self):
        self.enemies_killed += 1
        if self.enemies_killed >= cfg.TOTAL_ENEMIES:
            director.replace(EndScene())

    def on_found_a_bit(self, space, arbiter):
        cshape = arbiter.shapes[1]

        self.space.remove(cshape)
        cshape.sprite.kill()
        cshape.sprite = None

        self.active_room.cshapes.remove(cshape)
        self.bits[cshape.bit_index - 1].activate()

        return 0

    def on_enemy_shoot(self, enemy):
        self.create_bullet(enemy)

    def on_shooting_begin(self):
        self.schedule_interval(self.player_shoot, 1.0)
        self.player_shoot(0)

    def on_shooting_end(self):
        self.unschedule(self.player_shoot)

    def player_shoot(self, dt):
        if self.sensor.get_target() and self.player.is_sensor_on:
            self.create_bullet(self.player)

    def create_bullet(self, shooter):
        bullet = Bullet(shooter, self.on_bullet_out)

        self.game_objects.add(bullet)
        self.bullets.append(bullet)
        self.space.add(bullet.cshape)

    def on_bullet_out(self, bullet):
        self.bullets.remove(bullet)
        bullet.kill()
        self.space.remove(bullet.cshape)

    def on_bullet_hit_target(self, space, arbiter):
        shape = arbiter.shapes[1]
        self.sensor.targets.remove(shape)
        shape.on_hit(space)
        return 0
